import { __values, __assign, __awaiter, __generator, __extends, __read } from 'tslib';
import { inject } from '@angular/core';
import { AngularFirestore } from '@angular/fire/firestore';
import { runStoreAction, StoreActions, withTransaction, arrayAdd, arrayUpsert, arrayRemove, arrayUpdate } from '@datorama/akita';
import { isObservable, of, combineLatest, race } from 'rxjs';
import { map, switchMap, tap, takeWhile, distinctUntilChanged, filter, shareReplay, finalize } from 'rxjs/operators';
import { Router } from '@angular/router';
import { RouterQuery } from '@datorama/akita-ng-router-store';
import { AngularFireAuth } from '@angular/fire/auth';
import { auth } from 'firebase';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// Helper to retrieve the id and path of a document in the collection
/**
 * @param {?} options
 * @param {?=} collectionPath
 * @return {?}
 */
function getIdAndPath(options, collectionPath) {
    /** @type {?} */
    var path = '';
    /** @type {?} */
    var id = '';
    if (options['id']) {
        if (!collectionPath) {
            throw new Error('You should provide the colletion path with the id');
        }
        id = options['id'];
        path = collectionPath + "/" + id;
    }
    else if (options['path']) {
        path = options['path'];
        /** @type {?} */
        var part = path.split('/');
        if (path.length % 2 !== 0) {
            throw new Error("Path " + path + " doesn't look like a Firestore's document path");
        }
        id = part[part.length - 1];
    }
    else {
        throw new Error("You should provide either an \"id\" OR a \"path\".");
    }
    return { id: id, path: path };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Set the loading parameter of a specific store
 * @param {?} storeName
 * @param {?} loading
 * @return {?}
 */
function setLoading(storeName, loading) {
    runStoreAction(storeName, StoreActions.Update, {
        payload: {
            data: { loading: loading }
        }
    });
}
/**
 *
 * @param {?} storeName
 * @param {?} active
 * @return {?}
 */
function setActive(storeName, active) {
    runStoreAction(storeName, StoreActions.Update, {
        payload: {
            data: { active: active }
        }
    });
}
/**
 * Create or update one or several entities in the store
 * @param {?} storeName
 * @param {?} data
 * @return {?}
 */
function upsertStoreEntity(storeName, data) {
    /** @type {?} */
    var payload = { data: data };
    runStoreAction(storeName, StoreActions.UpsertEntities, { payload: payload });
}
/**
 * Remove one or several entities in the store
 * @param {?} storeName
 * @param {?} entityIds
 * @return {?}
 */
function removeStoreEntity(storeName, entityIds) {
    /** @type {?} */
    var payload = { entityIds: entityIds };
    runStoreAction(storeName, StoreActions.RemoveEntities, { payload: payload });
}
/**
 * Update one or several entities in the store
 * @param {?} storeName
 * @param {?} entityIds
 * @param {?} data
 * @return {?}
 */
function updateStoreEntity(storeName, entityIds, data) {
    /** @type {?} */
    var payload = {
        data: data,
        entityIds: entityIds
    };
    runStoreAction(storeName, StoreActions.UpdateEntities, { payload: payload });
}
/**
 * Sync a specific store with actions from Firestore
 * @template S
 * @param {?} storeName
 * @param {?} actions
 * @param {?=} idKey
 * @return {?}
 */
function syncStoreFromDocAction(storeName, actions, idKey) {
    var e_1, _a, _b;
    if (idKey === void 0) { idKey = 'id'; }
    setLoading(storeName, false);
    if (actions.length === 0) {
        return;
    }
    try {
        for (var actions_1 = __values(actions), actions_1_1 = actions_1.next(); !actions_1_1.done; actions_1_1 = actions_1.next()) {
            var action = actions_1_1.value;
            /** @type {?} */
            var id = action.payload.doc.id;
            /** @type {?} */
            var entity = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    upsertStoreEntity(storeName, __assign((_b = {}, _b[idKey] = id, _b), entity));
                    break;
                }
                case 'removed': {
                    removeStoreEntity(storeName, id);
                    break;
                }
                case 'modified': {
                    updateStoreEntity(storeName, id, entity);
                    break;
                }
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (actions_1_1 && !actions_1_1.done && (_a = actions_1.return)) _a.call(actions_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
}
/**
 * Sync a specific store with actions from Firestore
 * @template S
 * @param {?} storeName
 * @param {?} action
 * @param {?=} idKey
 * @return {?}
 */
function syncStoreFromDocActionSnapshot(storeName, action, idKey) {
    var _a;
    if (idKey === void 0) { idKey = 'id'; }
    setLoading(storeName, false);
    /** @type {?} */
    var id = action.payload.id;
    /** @type {?} */
    var entity = action.payload.data();
    if (!action.payload.exists) {
        removeStoreEntity(storeName, id);
    }
    else {
        upsertStoreEntity(storeName, __assign((_a = {}, _a[idKey] = id, _a), entity));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Get the store name of a store to be synced
 * @param {?} store
 * @param {?=} storeOptions
 * @return {?}
 */
function getStoreName(store, storeOptions) {
    if (storeOptions === void 0) { storeOptions = {}; }
    if (!store && !storeOptions.storeName) {
        throw new Error('You should either provide a store name or inject a store instance in constructor');
    }
    return storeOptions.storeName || store.storeName;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template S
 */
var  /**
 * @template S
 */
CollectionService = /** @class */ (function () {
    function CollectionService(store, collectionPath) {
        this.store = store;
        this.collectionPath = collectionPath;
        if (!this.constructor['path'] && !this.collectionPath) {
            throw new Error('You should provide a path to the collection');
        }
        try {
            this.db = inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionService requires AngularFirestore.');
        }
    }
    Object.defineProperty(CollectionService.prototype, "idKey", {
        get: /**
         * @return {?}
         */
        function () {
            return this.constructor['idKey']
                || this.store ? this.store.idKey : 'id';
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "path", {
        /** The path to the collection in Firestore */
        get: /**
         * The path to the collection in Firestore
         * @return {?}
         */
        function () {
            return this.constructor['path'] || this.collectionPath;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "currentPath", {
        /** A snapshot of the path */
        get: /**
         * A snapshot of the path
         * @return {?}
         */
        function () {
            if (isObservable(this.path)) {
                throw new Error('Cannot get a snapshot of the path if it is an Observable');
            }
            return this.path;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "path$", {
        /** An observable version of the path */
        get: /**
         * An observable version of the path
         * @return {?}
         */
        function () {
            return isObservable(this.path) ? this.path : of(this.path);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "collection", {
        /**
         * The Angular Fire collection
         * @notice If path is an observable, it becomes an observable.
         */
        get: /**
         * The Angular Fire collection
         * \@notice If path is an observable, it becomes an observable.
         * @return {?}
         */
        function () {
            return this.db.collection(this.currentPath);
        },
        enumerable: true,
        configurable: true
    });
    /** Preformat the document before updating Firestore */
    /**
     * Preformat the document before updating Firestore
     * @protected
     * @template E
     * @param {?} document
     * @return {?}
     */
    CollectionService.prototype.preFormat = /**
     * Preformat the document before updating Firestore
     * @protected
     * @template E
     * @param {?} document
     * @return {?}
     */
    function (document) {
        return document;
    };
    Object.defineProperty(CollectionService.prototype, "config", {
        /** The config given by the `CollectonConfig` */
        get: /**
         * The config given by the `CollectonConfig`
         * @return {?}
         */
        function () {
            return {
                path: this.constructor['path'],
                idKey: this.constructor['idKey']
            };
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param {?=} pathOrQuery
     * @param {?=} queryOrOptions
     * @param {?=} storeOptions
     * @return {?}
     */
    CollectionService.prototype.syncCollection = /**
     * @param {?=} pathOrQuery
     * @param {?=} queryOrOptions
     * @param {?=} storeOptions
     * @return {?}
     */
    function (pathOrQuery, queryOrOptions, storeOptions) {
        var _this = this;
        if (pathOrQuery === void 0) { pathOrQuery = this.path; }
        if (storeOptions === void 0) { storeOptions = { loading: true }; }
        /** @type {?} */
        var path$;
        /** @type {?} */
        var queryFn;
        // check type of pathOrQuery
        if (isObservable(pathOrQuery)) {
            path$ = pathOrQuery;
        }
        else if (typeof pathOrQuery === 'function') {
            queryFn = pathOrQuery;
            path$ = this.path$;
        }
        else if (typeof pathOrQuery === 'object') {
            storeOptions = pathOrQuery;
            path$ = this.path$;
        }
        else {
            path$ = this.path$;
        }
        // check type of queryOrOptions
        if (typeof queryOrOptions === 'function') {
            queryFn = queryOrOptions;
        }
        else if (typeof queryOrOptions === 'object') {
            storeOptions = queryOrOptions;
        }
        /** @type {?} */
        var storeName = getStoreName(this.store, storeOptions);
        if (storeOptions.loading) {
            setLoading(storeName, true);
        }
        // Start Listening
        return path$.pipe(map((/**
         * @param {?} collectionPath
         * @return {?}
         */
        function (collectionPath) { return _this.db.collection(collectionPath, queryFn); })), switchMap((/**
         * @param {?} collection
         * @return {?}
         */
        function (collection) { return collection.stateChanges(); })), withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        function (actions) { return syncStoreFromDocAction(storeName, actions, _this.idKey); })));
    };
    /**
     * @param {?=} idOrQuery
     * @param {?=} queryOrOption
     * @param {?=} storeOptions
     * @return {?}
     */
    CollectionService.prototype.syncCollectionGroup = /**
     * @param {?=} idOrQuery
     * @param {?=} queryOrOption
     * @param {?=} storeOptions
     * @return {?}
     */
    function (idOrQuery, queryOrOption, storeOptions) {
        var _this = this;
        if (idOrQuery === void 0) { idOrQuery = this.currentPath; }
        if (storeOptions === void 0) { storeOptions = { loading: true }; }
        /** @type {?} */
        var path;
        /** @type {?} */
        var query;
        if (typeof idOrQuery === 'string') {
            path = idOrQuery;
        }
        else if (typeof idOrQuery === 'function') {
            path = this.currentPath;
            query = idOrQuery;
        }
        else if (typeof idOrQuery === 'object') {
            path = this.currentPath;
            storeOptions = idOrQuery;
        }
        else {
            throw new Error('1ier parameter if either a string, a queryFn or a StoreOption');
        }
        if (typeof queryOrOption === 'function') {
            query = queryOrOption;
        }
        else if (typeof queryOrOption === 'object') {
            storeOptions = queryOrOption;
        }
        /** @type {?} */
        var storeName = getStoreName(this.store, storeOptions);
        if (storeOptions.loading) {
            setLoading(storeName, true);
        }
        /** @type {?} */
        var collectionId = path.split('/').pop();
        return this.db.collectionGroup(collectionId, query).stateChanges().pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        function (actions) { return syncStoreFromDocAction(storeName, actions, _this.idKey); })));
    };
    /**
     * Sync the store with several documents
     * @param ids$ An array of ids or an observable of ids
     * @param storeOptions Options on the store to sync to
     */
    /**
     * Sync the store with several documents
     * @param {?} ids$ An array of ids or an observable of ids
     * @param {?=} storeOptions Options on the store to sync to
     * @return {?}
     */
    CollectionService.prototype.syncManyDocs = /**
     * Sync the store with several documents
     * @param {?} ids$ An array of ids or an observable of ids
     * @param {?=} storeOptions Options on the store to sync to
     * @return {?}
     */
    function (ids$, storeOptions) {
        var _this = this;
        if (storeOptions === void 0) { storeOptions = { loading: true }; }
        if (!isObservable(ids$)) {
            ids$ = of(ids$);
        }
        /** @type {?} */
        var storeName = getStoreName(this.store, storeOptions);
        if (storeOptions.loading) {
            setLoading(storeName, true);
        }
        return ids$.pipe(switchMap(((/**
         * @param {?} ids
         * @return {?}
         */
        function (ids) {
            if (!ids.length) {
                return of([]);
            }
            /** @type {?} */
            var syncs = ids.map((/**
             * @param {?} id
             * @return {?}
             */
            function (id) { return _this.path$.pipe(map((/**
             * @param {?} collectionPath
             * @return {?}
             */
            function (collectionPath) { return getIdAndPath({ id: id }, collectionPath); })), switchMap((/**
             * @param {?} __0
             * @return {?}
             */
            function (_a) {
                var path = _a.path;
                return _this.db.doc(path).snapshotChanges();
            })), map((/**
             * @param {?} action
             * @return {?}
             */
            function (action) { return syncStoreFromDocActionSnapshot(storeName, action, _this.idKey); }))); }));
            return combineLatest(syncs);
        }))));
    };
    /**
     * Stay in sync with one document
     * @param docOptions An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     * @param storeOptions Options on the store to sync to
     */
    /**
     * Stay in sync with one document
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} docOptions An object with EITHER `id` OR `path`.
     * @param {?=} storeOptions Options on the store to sync to
     * @return {?}
     */
    CollectionService.prototype.syncDoc = /**
     * Stay in sync with one document
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} docOptions An object with EITHER `id` OR `path`.
     * @param {?=} storeOptions Options on the store to sync to
     * @return {?}
     */
    function (docOptions, storeOptions) {
        var _this = this;
        if (storeOptions === void 0) { storeOptions = { loading: false }; }
        /** @type {?} */
        var id;
        /** @type {?} */
        var storeName = getStoreName(this.store, storeOptions);
        if (storeOptions.loading) {
            setLoading(storeName, true);
        }
        return this.path$.pipe(map((/**
         * @param {?} collectionPath
         * @return {?}
         */
        function (collectionPath) { return getIdAndPath(docOptions, collectionPath); })), tap((/**
         * @param {?} data
         * @return {?}
         */
        function (data) {
            id = (/** @type {?} */ (data.id));
            if (storeOptions.loading) {
                setLoading(storeName, true);
            }
        })), switchMap((/**
         * @param {?} __0
         * @return {?}
         */
        function (_a) {
            var path = _a.path;
            return _this.db.doc(path).valueChanges();
        })), map((/**
         * @param {?} entity
         * @return {?}
         */
        function (entity) {
            var _a;
            /** @type {?} */
            var data = __assign((_a = {}, _a[_this.idKey] = id, _a), entity);
            upsertStoreEntity(storeName, data);
            setLoading(storeName, false);
            return data;
        })));
    };
    /**
     * @param {?} options
     * @param {?=} storeOptions
     * @return {?}
     */
    CollectionService.prototype.syncActive = /**
     * @param {?} options
     * @param {?=} storeOptions
     * @return {?}
     */
    function (options, storeOptions) {
        var _this = this;
        /** @type {?} */
        var storeName = getStoreName(this.store, storeOptions);
        if (Array.isArray(options)) {
            return this.syncManyDocs(options, storeOptions).pipe(tap((/**
             * @param {?} _
             * @return {?}
             */
            function (_) { return setActive(storeName, options); })));
        }
        else {
            return this.syncDoc(options, storeOptions).pipe(tap((/**
             * @param {?} entity
             * @return {?}
             */
            function (entity) { return setActive(storeName, entity[_this.idKey]); })));
        }
    };
    /**
     * @param {?=} idOrQuery
     * @return {?}
     */
    CollectionService.prototype.getValue = /**
     * @param {?=} idOrQuery
     * @return {?}
     */
    function (idOrQuery) {
        return __awaiter(this, void 0, void 0, function () {
            var _a, snapshot, snapshot;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        if (!(typeof idOrQuery === 'string')) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.db.doc(this.currentPath + "/" + idOrQuery).ref.get()];
                    case 1:
                        snapshot = _b.sent();
                        return [2 /*return*/, (/** @type {?} */ (__assign({}, snapshot.data(), (_a = {}, _a[this.idKey] = snapshot.id, _a))))];
                    case 2: return [4 /*yield*/, this.db.collection(this.currentPath, idOrQuery).ref.get()];
                    case 3:
                        snapshot = _b.sent();
                        return [2 /*return*/, snapshot.docs.map((/**
                             * @param {?} doc
                             * @return {?}
                             */
                            function (doc) {
                                var _a;
                                return (/** @type {?} */ ((__assign({}, doc.data(), (_a = {}, _a[_this.idKey] = doc.id, _a)))));
                            }))];
                }
            });
        });
    };
    /**
     * Add a document or a list of document to Firestore
     * @param docs A document or a list of document
     * @param write batch or transaction to run the operation into
     */
    /**
     * Add a document or a list of document to Firestore
     * @param {?} documents
     * @param {?=} options
     * @return {?}
     */
    CollectionService.prototype.add = /**
     * Add a document or a list of document to Firestore
     * @param {?} documents
     * @param {?=} options
     * @return {?}
     */
    function (documents, options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var docs, _a, write, ctx, operations, ids;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        docs = Array.isArray(documents) ? documents : [documents];
                        _a = options.write, write = _a === void 0 ? this.db.firestore.batch() : _a, ctx = options.ctx;
                        operations = docs.map((/**
                         * @param {?} doc
                         * @return {?}
                         */
                        function (doc) { return __awaiter(_this, void 0, void 0, function () {
                            var _a, id, data, ref;
                            return __generator(this, function (_b) {
                                switch (_b.label) {
                                    case 0:
                                        id = doc[this.idKey] || this.db.createId();
                                        data = this.preFormat(__assign({}, doc, (_a = {}, _a[this.idKey] = id, _a)));
                                        ref = this.db.doc(this.currentPath + "/" + id).ref;
                                        write.set(ref, data);
                                        if (!this.onCreate) return [3 /*break*/, 2];
                                        return [4 /*yield*/, this.onCreate(data, { write: write, ctx: ctx })];
                                    case 1:
                                        _b.sent();
                                        _b.label = 2;
                                    case 2: return [2 /*return*/, id];
                                }
                            });
                        }); }));
                        return [4 /*yield*/, Promise.all(operations)];
                    case 1:
                        ids = _b.sent();
                        // If there is no atomic write provided
                        if (!options.write) {
                            return [2 /*return*/, ((/** @type {?} */ (write))).commit()];
                        }
                        return [2 /*return*/, Array.isArray(documents) ? ids : ids[0]];
                }
            });
        });
    };
    /**
     * Remove one or several document from Firestore
     * @param id A unique or list of id representing the document
     * @param write batch or transaction to run the operation into
     */
    /**
     * Remove one or several document from Firestore
     * @param {?} id A unique or list of id representing the document
     * @param {?=} options
     * @return {?}
     */
    CollectionService.prototype.remove = /**
     * Remove one or several document from Firestore
     * @param {?} id A unique or list of id representing the document
     * @param {?=} options
     * @return {?}
     */
    function (id, options) {
        return __awaiter(this, void 0, void 0, function () {
            var ids, _a, write, ctx, operations;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        ids = Array.isArray(id) ? id : [id];
                        _a = options.write, write = _a === void 0 ? this.db.firestore.batch() : _a, ctx = options.ctx;
                        operations = ids.map((/**
                         * @param {?} docId
                         * @return {?}
                         */
                        function (docId) { return __awaiter(_this, void 0, void 0, function () {
                            var ref;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        ref = this.db.doc(this.currentPath + "/" + docId).ref;
                                        write.delete(ref);
                                        if (!this.onDelete) return [3 /*break*/, 2];
                                        return [4 /*yield*/, this.onDelete(docId, { write: write, ctx: ctx })];
                                    case 1:
                                        _a.sent();
                                        _a.label = 2;
                                    case 2: return [2 /*return*/];
                                }
                            });
                        }); }));
                        return [4 /*yield*/, Promise.all(operations)];
                    case 1:
                        _b.sent();
                        // If there is no atomic write provided
                        if (!options.write) {
                            return [2 /*return*/, ((/** @type {?} */ (write))).commit()];
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * @param {?} idsOrEntity
     * @param {?=} stateFnOrWrite
     * @param {?=} options
     * @return {?}
     */
    CollectionService.prototype.update = /**
     * @param {?} idsOrEntity
     * @param {?=} stateFnOrWrite
     * @param {?=} options
     * @return {?}
     */
    function (idsOrEntity, stateFnOrWrite, options) {
        return __awaiter(this, void 0, void 0, function () {
            var ids, newStateOrFn, isEntity, _a, write_1, ctx_1, operations;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        ids = [];
                        newStateOrFn = (/** @type {?} */ (stateFnOrWrite));
                        isEntity = (/**
                         * @param {?} value
                         * @return {?}
                         */
                        function (value) {
                            return typeof value === 'object' && idsOrEntity[_this.idKey];
                        });
                        if (isEntity(idsOrEntity)) {
                            ids = [idsOrEntity[this.idKey]];
                            options = (/** @type {?} */ (stateFnOrWrite));
                            newStateOrFn = idsOrEntity;
                        }
                        else {
                            ids = Array.isArray(idsOrEntity) ? idsOrEntity : [idsOrEntity];
                        }
                        if (!Array.isArray(ids) || !ids.length) {
                            return [2 /*return*/];
                        }
                        // If update depends on the entity, use transaction
                        if (typeof newStateOrFn === 'function') {
                            return [2 /*return*/, this.db.firestore.runTransaction((/**
                                 * @param {?} tx
                                 * @return {?}
                                 */
                                function (tx) { return __awaiter(_this, void 0, void 0, function () {
                                    var operations;
                                    var _this = this;
                                    return __generator(this, function (_a) {
                                        operations = ids.map((/**
                                         * @param {?} id
                                         * @return {?}
                                         */
                                        function (id) { return __awaiter(_this, void 0, void 0, function () {
                                            var _a, ref, snapshot, doc, data;
                                            return __generator(this, function (_b) {
                                                switch (_b.label) {
                                                    case 0:
                                                        ref = this.db.doc(this.currentPath + "/" + id).ref;
                                                        return [4 /*yield*/, tx.get(ref)];
                                                    case 1:
                                                        snapshot = _b.sent();
                                                        doc = Object.freeze((/** @type {?} */ (__assign({}, snapshot.data, (_a = {}, _a[this.idKey] = id, _a)))));
                                                        data = ((/** @type {?} */ (newStateOrFn)))(this.preFormat(doc));
                                                        tx.update(ref, data);
                                                        if (!this.onUpdate) return [3 /*break*/, 3];
                                                        return [4 /*yield*/, this.onUpdate(data, { write: tx, ctx: options.ctx })];
                                                    case 2:
                                                        _b.sent();
                                                        _b.label = 3;
                                                    case 3: return [2 /*return*/, tx];
                                                }
                                            });
                                        }); }));
                                        return [2 /*return*/, Promise.all(operations)];
                                    });
                                }); }))];
                        }
                        if (!isEntity(newStateOrFn)) return [3 /*break*/, 2];
                        _a = options.write, write_1 = _a === void 0 ? this.db.firestore.batch() : _a, ctx_1 = options.ctx;
                        operations = ids.map((/**
                         * @param {?} docId
                         * @return {?}
                         */
                        function (docId) { return __awaiter(_this, void 0, void 0, function () {
                            var doc, data, ref;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        doc = Object.freeze((/** @type {?} */ (newStateOrFn)));
                                        data = this.preFormat(doc);
                                        ref = this.db.doc(this.currentPath + "/" + docId).ref;
                                        write_1.update(ref, data);
                                        if (!this.onUpdate) return [3 /*break*/, 2];
                                        return [4 /*yield*/, this.onUpdate(data, { write: write_1, ctx: ctx_1 })];
                                    case 1:
                                        _a.sent();
                                        _a.label = 2;
                                    case 2: return [2 /*return*/];
                                }
                            });
                        }); }));
                        return [4 /*yield*/, Promise.all(operations)];
                    case 1:
                        _b.sent();
                        // If there is no atomic write provided
                        if (!options.write) {
                            return [2 /*return*/, ((/** @type {?} */ (write_1))).commit()];
                        }
                        _b.label = 2;
                    case 2: return [2 /*return*/];
                }
            });
        });
    };
    return CollectionService;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Set the configuration for the collection service
 * @param {?=} options
 * @return {?}
 */
function CollectionConfig(options) {
    if (options === void 0) { options = {}; }
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    function (constructor) {
        Object.keys(options).forEach((/**
         * @param {?} key
         * @return {?}
         */
        function (key) { return constructor[key] = options[key]; }));
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} data
 * @return {?}
 */
function CollectionGuardConfig(data) {
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    function (constructor) {
        Object.keys(data).forEach((/**
         * @param {?} key
         * @return {?}
         */
        function (key) { return (constructor[key] = data[key]); }));
    });
}
/**
 * @template S
 */
var  /**
 * @template S
 */
CollectionGuard = /** @class */ (function () {
    function CollectionGuard(service) {
        this.service = service;
        try {
            this.router = inject(Router);
        }
        catch (err) {
            throw new Error('CollectionGuard requires RouterModule to be imported');
        }
    }
    Object.defineProperty(CollectionGuard.prototype, "awaitSync", {
        // Can be override by the extended class
        /** Should the guard wait for connection to Firestore to complete */
        get: 
        // Can be override by the extended class
        /**
         * Should the guard wait for connection to Firestore to complete
         * @protected
         * @return {?}
         */
        function () {
            return this.constructor['awaitSync'] || false;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionGuard.prototype, "queryFn", {
        // Can be override by the extended class
        /** A queryFn to execute for sync */
        get: 
        // Can be override by the extended class
        /**
         * A queryFn to execute for sync
         * @protected
         * @return {?}
         */
        function () {
            return this.constructor['queryFn'];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionGuard.prototype, "redirect", {
        // Can be override by the extended class
        /** The route to redirect to if you sync failed */
        get: 
        // Can be override by the extended class
        /**
         * The route to redirect to if you sync failed
         * @protected
         * @return {?}
         */
        function () {
            return this.constructor['redirect'];
        },
        enumerable: true,
        configurable: true
    });
    // Can be override by the extended class
    /** The method to subscribe to while route is active */
    // Can be override by the extended class
    /**
     * The method to subscribe to while route is active
     * @protected
     * @param {?} next
     * @return {?}
     */
    CollectionGuard.prototype.sync = 
    // Can be override by the extended class
    /**
     * The method to subscribe to while route is active
     * @protected
     * @param {?} next
     * @return {?}
     */
    function (next) {
        var _a = (/** @type {?} */ (next.data)).queryFn, queryFn = _a === void 0 ? this.queryFn : _a;
        return this.service.syncCollection(queryFn);
    };
    /**
     * @param {?} next
     * @return {?}
     */
    CollectionGuard.prototype.canActivate = /**
     * @param {?} next
     * @return {?}
     */
    function (next) {
        var _this = this;
        var _a = (/** @type {?} */ (next.data)), _b = _a.redirect, redirect = _b === void 0 ? this.redirect : _b, _c = _a.awaitSync, awaitSync = _c === void 0 ? this.awaitSync : _c;
        return new Promise((/**
         * @param {?} res
         * @param {?} rej
         * @return {?}
         */
        function (res, rej) {
            if (awaitSync) {
                _this.subscription = _this.sync(next).pipe(takeWhile((/**
                 * @param {?} result
                 * @return {?}
                 */
                function (result) { return typeof result !== 'string'; }))).subscribe({
                    next: (/**
                     * @param {?} result
                     * @return {?}
                     */
                    function (result) {
                        switch (typeof result) {
                            case 'string':
                                return res(_this.router.parseUrl(result));
                            case 'boolean':
                                return res(result);
                            default:
                                return res(true);
                        }
                    }),
                    error: (/**
                     * @param {?} err
                     * @return {?}
                     */
                    function (err) {
                        res(_this.router.parseUrl(redirect || ''));
                        throw new Error(err);
                    })
                });
            }
            else {
                _this.subscription = _this.sync(next).subscribe();
                res(true);
            }
        }));
    };
    /**
     * @return {?}
     */
    CollectionGuard.prototype.canDeactivate = /**
     * @return {?}
     */
    function () {
        if (this.subscription) {
            this.subscription.unsubscribe();
        }
        return true;
    };
    return CollectionGuard;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Get the params from a path
 * @param {?} path
 * @return {?}
 */
function getPathParams(path) {
    return path.split('/')
        .filter((/**
     * @param {?} segment
     * @return {?}
     */
    function (segment) { return segment.charAt(0) === ':'; }))
        .map((/**
     * @param {?} segment
     * @return {?}
     */
    function (segment) { return segment.substr(1); }));
}
/**
 * Transform a path based on the params
 * \@example pathWithParams('movies/:movieId/stakeholder/:shId', { movieId, shId })
 * @param {?} path The path with params starting with "/:"
 * @param {?} params A map of id params
 * @return {?}
 */
function pathWithParams(path, params) {
    return path.split('/').map((/**
     * @param {?} segment
     * @return {?}
     */
    function (segment) {
        if (segment.charAt(0) === ':') {
            /** @type {?} */
            var key = segment.substr(1);
            if (!params[key]) {
                throw new Error("Required parameter " + key + " from " + path + " doesn't exist in params " + JSON.stringify(params));
            }
            return params[key];
        }
        else {
            return segment;
        }
    })).join('/');
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template S
 */
var  /**
 * @template S
 */
SubcollectionService = /** @class */ (function (_super) {
    __extends(SubcollectionService, _super);
    function SubcollectionService(store) {
        var _this = _super.call(this, store) || this;
        _this.store = store;
        try {
            _this.routerQuery = inject(RouterQuery);
        }
        catch (err) {
            throw new Error('SubCollectionService requires a "AkitaNgRouterStoreModule" from Akita.');
        }
        return _this;
    }
    Object.defineProperty(SubcollectionService.prototype, "path", {
        get: /**
         * @return {?}
         */
        function () {
            var _this = this;
            /** @type {?} */
            var pathParams = getPathParams(this.constructor['path']);
            return this.routerQuery.selectParams().pipe(
            // Don't trigger change if params needed (and only them) haven't changed
            distinctUntilChanged((/**
             * @param {?} old
             * @param {?} next
             * @return {?}
             */
            function (old, next) {
                return !pathParams.find((/**
                 * @param {?} param
                 * @return {?}
                 */
                function (param) { return old[param] !== next[param]; }));
            })), 
            // Need to filter because changes in params comes before canDeactivate
            filter((/**
             * @param {?} params
             * @return {?}
             */
            function (params) { return pathParams.every((/**
             * @param {?} param
             * @return {?}
             */
            function (param) { return !!params[param]; })); })), map((/**
             * @param {?} params
             * @return {?}
             */
            function (params) { return pathWithParams(_this.constructor['path'], params); })), tap((/**
             * @param {?} path
             * @return {?}
             */
            function (path) {
                if (path !== _this.pathToCollection) {
                    _this.store.reset();
                    _this.pathToCollection = path;
                }
            })), shareReplay(1));
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SubcollectionService.prototype, "currentPath", {
        get: /**
         * @return {?}
         */
        function () {
            /** @type {?} */
            var params = this.routerQuery.getParams();
            return pathWithParams(this.constructor['path'], params);
        },
        enumerable: true,
        configurable: true
    });
    return SubcollectionService;
}(CollectionService));

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @abstract
 * @template S
 */
var  /**
 * @abstract
 * @template S
 */
CollectionGroupService = /** @class */ (function () {
    function CollectionGroupService(store) {
        this.store = store;
        try {
            this.db = inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionGroupService requires AngularFirestore.');
        }
    }
    Object.defineProperty(CollectionGroupService.prototype, "idKey", {
        get: /**
         * @return {?}
         */
        function () {
            if (this.store) {
                return this.store.idKey;
            }
        },
        enumerable: true,
        configurable: true
    });
    /**
     * @param {?=} queryOrOptions
     * @param {?=} storeOptions
     * @return {?}
     */
    CollectionGroupService.prototype.syncCollection = /**
     * @param {?=} queryOrOptions
     * @param {?=} storeOptions
     * @return {?}
     */
    function (queryOrOptions, storeOptions) {
        var _this = this;
        if (storeOptions === void 0) { storeOptions = { loading: true }; }
        /** @type {?} */
        var query;
        if (typeof queryOrOptions === 'function') {
            query = queryOrOptions;
        }
        else if (typeof queryOrOptions === 'object') {
            storeOptions = queryOrOptions;
        }
        /** @type {?} */
        var storeName = getStoreName(this.store, storeOptions);
        if (storeOptions.loading) {
            setLoading(storeName, true);
        }
        return this.db.collectionGroup(this.collectionId, query).stateChanges().pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        function (actions) { return syncStoreFromDocAction(storeName, actions, _this.idKey); })));
    };
    /** Return a snapshot of the collection group */
    /**
     * Return a snapshot of the collection group
     * @param {?=} queryGroupFn
     * @return {?}
     */
    CollectionGroupService.prototype.getValue = /**
     * Return a snapshot of the collection group
     * @param {?=} queryGroupFn
     * @return {?}
     */
    function (queryGroupFn) {
        return __awaiter(this, void 0, void 0, function () {
            var snapshot;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.db.collectionGroup(this.collectionId, queryGroupFn).get().toPromise()];
                    case 1:
                        snapshot = _a.sent();
                        return [2 /*return*/, snapshot.docs.map((/**
                             * @param {?} doc
                             * @return {?}
                             */
                            function (doc) { return (/** @type {?} */ (doc.data())); }))];
                }
            });
        });
    };
    return CollectionGroupService;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
var fireAuthProviders = (/** @type {?} */ (['github', 'google', 'microsoft', 'facebook', 'twitter', 'email']));
/**
 * @param {?} provider
 * @return {?}
 */
function isFireAuthProvider(provider) {
    return fireAuthProviders.includes((/** @type {?} */ (provider)));
}
/**
 * @param {?} provider
 * @return {?}
 */
function getAuthProvider(provider) {
    switch (provider) {
        case 'email': return new auth.EmailAuthProvider();
        case 'facebook': return new auth.FacebookAuthProvider();
        case 'github': return new auth.GithubAuthProvider();
        case 'google': return new auth.GoogleAuthProvider();
        case 'microsoft': return new auth.OAuthProvider('microsoft.com');
        case 'twitter': return new auth.TwitterAuthProvider();
    }
}
/**
 * @template S
 */
var  /**
 * @template S
 */
FireAuthService = /** @class */ (function () {
    function FireAuthService(store) {
        this.store = store;
        this.collectionPath = 'users';
        this.db = inject(AngularFirestore);
        this.fireAuth = inject(AngularFireAuth);
        this.collection = this.db.collection(this.path);
    }
    /** Can be overrided */
    /**
     * Can be overrided
     * @protected
     * @param {?} user
     * @return {?}
     */
    FireAuthService.prototype.selectProfile = /**
     * Can be overrided
     * @protected
     * @param {?} user
     * @return {?}
     */
    function (user) {
        return this.collection.doc(user.uid).valueChanges();
    };
    /** Can be overrided */
    /**
     * Can be overrided
     * @protected
     * @param {?} user
     * @return {?}
     */
    FireAuthService.prototype.selectRoles = /**
     * Can be overrided
     * @protected
     * @param {?} user
     * @return {?}
     */
    function (user) {
        return user.getIdTokenResult().then((/**
         * @param {?} __0
         * @return {?}
         */
        function (_a) {
            var claims = _a.claims;
            return (/** @type {?} */ (claims));
        }));
    };
    /** Should be override */
    /**
     * Should be override
     * @protected
     * @param {?} user
     * @return {?}
     */
    FireAuthService.prototype.createProfile = /**
     * Should be override
     * @protected
     * @param {?} user
     * @return {?}
     */
    function (user) {
        return (/** @type {?} */ ({
            photoURL: user.photoURL,
            displayName: user.displayName,
        }));
    };
    Object.defineProperty(FireAuthService.prototype, "user", {
        /** The current sign-in user (or null) */
        get: /**
         * The current sign-in user (or null)
         * @return {?}
         */
        function () {
            return this.fireAuth.auth.currentUser;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(FireAuthService.prototype, "path", {
        /** The path to the profile in firestore */
        get: /**
         * The path to the profile in firestore
         * @return {?}
         */
        function () {
            return this.constructor['path'] || this.collectionPath;
        },
        enumerable: true,
        configurable: true
    });
    /** Start listening on User */
    /**
     * Start listening on User
     * @return {?}
     */
    FireAuthService.prototype.sync = /**
     * Start listening on User
     * @return {?}
     */
    function () {
        var _this = this;
        return this.fireAuth.authState.pipe(switchMap((/**
         * @param {?} user
         * @return {?}
         */
        function (user) { return user ? combineLatest([
            of(user.uid),
            _this.selectProfile(user),
            _this.selectRoles(user),
        ]) : of([null, null, null]); })), tap((/**
         * @param {?} __0
         * @return {?}
         */
        function (_a) {
            var _b = __read(_a, 3), uid = _b[0], profile = _b[1], roles = _b[2];
            return _this.store.update((/** @type {?} */ ({ uid: uid, profile: profile, roles: roles })));
        })));
    };
    /**
     * @description Delete user from authentication service and database
     * WARNING This is security sensitive operation
     */
    /**
     * \@description Delete user from authentication service and database
     * WARNING This is security sensitive operation
     * @return {?}
     */
    FireAuthService.prototype.delete = /**
     * \@description Delete user from authentication service and database
     * WARNING This is security sensitive operation
     * @return {?}
     */
    function () {
        return __awaiter(this, void 0, void 0, function () {
            var batch, ref;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!this.user) {
                            throw new Error('No user connected');
                        }
                        batch = this.db.firestore.batch();
                        ref = this.collection.doc(this.user.uid).ref;
                        batch.delete(ref);
                        if (!this.onDelete) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.onDelete(batch)];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2: return [4 /*yield*/, batch.commit()];
                    case 3:
                        _a.sent();
                        return [2 /*return*/, this.user.delete()];
                }
            });
        });
    };
    /** Update the current profile of the authenticated user */
    /**
     * Update the current profile of the authenticated user
     * @param {?} profile
     * @return {?}
     */
    FireAuthService.prototype.update = /**
     * Update the current profile of the authenticated user
     * @param {?} profile
     * @return {?}
     */
    function (profile) {
        return __awaiter(this, void 0, void 0, function () {
            var batch, ref;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!this.user.uid) {
                            throw new Error('No user connected.');
                        }
                        batch = this.db.firestore.batch();
                        ref = this.collection.doc(this.user.uid).ref;
                        batch.update(ref, profile);
                        if (!this.onCreate) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.onCreate(profile, batch)];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2: return [2 /*return*/, batch.commit()];
                }
            });
        });
    };
    /** Create a user based on email and password */
    /**
     * Create a user based on email and password
     * @param {?} email
     * @param {?} password
     * @return {?}
     */
    FireAuthService.prototype.signup = /**
     * Create a user based on email and password
     * @param {?} email
     * @param {?} password
     * @return {?}
     */
    function (email, password) {
        return __awaiter(this, void 0, void 0, function () {
            var cred, profile, batch, ref;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.fireAuth.auth.createUserWithEmailAndPassword(email, password)];
                    case 1:
                        cred = _a.sent();
                        return [4 /*yield*/, this.createProfile(cred.user)];
                    case 2:
                        profile = _a.sent();
                        batch = this.db.firestore.batch();
                        ref = this.collection.doc(cred.user.uid).ref;
                        batch.set(ref, profile);
                        if (this.onCreate) {
                            this.onCreate(profile, batch);
                        }
                        return [2 /*return*/, cred];
                }
            });
        });
    };
    /**
     * @param {?=} provider
     * @param {?=} password
     * @return {?}
     */
    FireAuthService.prototype.signin = /**
     * @param {?=} provider
     * @param {?=} password
     * @return {?}
     */
    function (provider, password) {
        return __awaiter(this, void 0, void 0, function () {
            var cred, authProvider, profile, batch, ref, err_1;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        this.store.setLoading(true);
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 14, , 15]);
                        cred = void 0;
                        if (!!provider) return [3 /*break*/, 3];
                        return [4 /*yield*/, this.fireAuth.auth.signInAnonymously()];
                    case 2:
                        cred = _a.sent();
                        return [3 /*break*/, 9];
                    case 3:
                        if (!password) return [3 /*break*/, 5];
                        return [4 /*yield*/, this.fireAuth.auth.signInWithEmailAndPassword(provider, password)];
                    case 4:
                        cred = _a.sent();
                        return [3 /*break*/, 9];
                    case 5:
                        if (!isFireAuthProvider(provider)) return [3 /*break*/, 7];
                        authProvider = getAuthProvider(provider);
                        return [4 /*yield*/, this.fireAuth.auth.signInWithPopup(authProvider)];
                    case 6:
                        cred = _a.sent();
                        return [3 /*break*/, 9];
                    case 7: return [4 /*yield*/, this.fireAuth.auth.signInWithCustomToken(provider)];
                    case 8:
                        cred = _a.sent();
                        _a.label = 9;
                    case 9:
                        if (!cred.additionalUserInfo.isNewUser) return [3 /*break*/, 13];
                        return [4 /*yield*/, this.createProfile(cred.user)];
                    case 10:
                        profile = _a.sent();
                        batch = this.db.firestore.batch();
                        ref = this.collection.doc(cred.user.uid).ref;
                        batch.set(ref, profile);
                        if (!this.onCreate) return [3 /*break*/, 12];
                        return [4 /*yield*/, this.onCreate(profile, batch)];
                    case 11:
                        _a.sent();
                        _a.label = 12;
                    case 12:
                        batch.commit();
                        _a.label = 13;
                    case 13:
                        this.store.setLoading(false);
                        return [2 /*return*/, cred];
                    case 14:
                        err_1 = _a.sent();
                        this.store.setLoading(false);
                        if (err_1.code === 'auth/operation-not-allowed') {
                            console.warn('You tried to connect with unenabled auth provider. Enable it in Firebase console');
                        }
                        throw err_1;
                    case 15: return [2 /*return*/];
                }
            });
        });
    };
    /** Signs out the current user */
    /**
     * Signs out the current user
     * @return {?}
     */
    FireAuthService.prototype.signOut = /**
     * Signs out the current user
     * @return {?}
     */
    function () {
        return this.fireAuth.auth.signOut();
    };
    return FireAuthService;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
var initialAuthState = {
    uid: null,
    profile: null,
    loading: false
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} write
 * @param {?} roles
 * @return {?}
 */
function canWrite(write, roles) {
    return roles.some((/**
     * @param {?} role
     * @return {?}
     */
    function (role) { return role === write || role === 'write'; }));
}
/**
 * @param {?} read
 * @param {?} roles
 * @return {?}
 */
function canRead(read, roles) {
    return roles.some((/**
     * @param {?} role
     * @return {?}
     */
    function (role) { return role === read || role === 'write'; }));
}
/**
 * @param {?} role
 * @param {?} roles
 * @return {?}
 */
function hasRole(role, roles) {
    switch (role) {
        case 'write': return roles.includes('write');
        case 'create': return canWrite('create', roles);
        case 'delete': return canWrite('delete', roles);
        case 'update': return canWrite('update', roles);
        case 'read': return roles.includes('read');
        case 'get': return canRead('get', roles);
        case 'list': return canRead('list', roles);
        default: return false;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} __0
 * @return {?}
 */
function shoudCancel(_a) {
    var validate = _a.validate, cancel = _a.cancel;
    return race([validate.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        function (_) { return false; }))), cancel.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        function (_) { return true; })))]);
}
/**
 * @param {?} __0
 * @return {?}
 */
function waitForCancel(_a) {
    var startWith = _a.startWith, endWith = _a.endWith, shouldValidate = _a.shouldValidate, shouldCancel = _a.shouldCancel;
    return __awaiter(this, void 0, void 0, function () {
        var cancelled;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    startWith();
                    return [4 /*yield*/, race([
                            shouldValidate.pipe(map((/**
                             * @param {?} _
                             * @return {?}
                             */
                            function (_) { return false; }))),
                            shouldCancel.pipe(map((/**
                             * @param {?} _
                             * @return {?}
                             */
                            function (_) { return true; })))
                        ]).toPromise()];
                case 1:
                    cancelled = _b.sent();
                    endWith(cancelled);
                    return [2 /*return*/];
            }
        });
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template E, K
 * @param {?} query
 * @param {?} parent
 * @return {?}
 */
function getSubQuery(query, parent) {
    if (typeof query !== 'function') {
        return query;
    }
    return query(parent);
}
/**
 * @param {?} path
 * @return {?}
 */
function isDocPath(path) {
    return path.split('/').length % 2 === 0;
}
/**
 * Transform a path into a collection Query
 * @template T
 * @param {?} path
 * @param {?=} queryFn
 * @return {?}
 */
function collection(path, queryFn) {
    return (/** @type {?} */ ({ path: path, queryFn: queryFn }));
}
/**
 * Transform a path into a doc query
 * @template T
 * @param {?} path
 * @return {?}
 */
function doc(path) {
    return (/** @type {?} */ ({ path: path }));
}
/**
 * Check if a value is a query
 * @template T
 * @param {?} query
 * @return {?}
 */
function isQuery(query) {
    if (typeof query === 'object') {
        return !!query['path'];
    }
    return false;
}
/**
 * Sync the collection
 * @template E
 * @this {?}
 * @param {?} query The query to trigger
 * @return {?}
 */
function syncQuery(query) {
    var _this = this;
    // If single query
    if (typeof query === 'string') {
        return isDocPath(query)
            ? this.syncDoc({ path: query })
            : this.syncCollection(query);
    }
    if (Array.isArray(query)) {
        return combineLatest(query.map((/**
         * @param {?} oneQuery
         * @return {?}
         */
        function (oneQuery) { return syncQuery.bind(_this, oneQuery); })));
    }
    if (!isQuery(query)) {
        throw new Error('Query should be either a path, a Query object or an array of Queries');
    }
    ////////////////
    // SUBQUERIES //
    ////////////////
    /**
     * Listen on Child actions
     * @type {?}
     */
    var fromChildAction = (/**
     * @param {?} actions
     * @param {?} child
     * @return {?}
     */
    function (actions, child) {
        var e_1, _a;
        /** @type {?} */
        var idKey = 'id';
        // TODO: Improve how to
        var parentId = child.parentId, key = child.key;
        var _loop_1 = function (action) {
            /** @type {?} */
            var id = action.payload.doc.id;
            /** @type {?} */
            var data = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    _this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    function (entity) {
                        var _a;
                        return (/** @type {?} */ ((_a = {},
                            _a[key] = arrayUpsert((/** @type {?} */ (entity[key])) || [], id, data, idKey),
                            _a)));
                    }));
                    break;
                }
                case 'removed': {
                    _this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    function (entity) {
                        var _a;
                        return (/** @type {?} */ ((_a = {},
                            _a[key] = arrayRemove((/** @type {?} */ (entity[key])), id, idKey),
                            _a)));
                    }));
                    break;
                }
                case 'modified': {
                    _this['store'].update((/** @type {?} */ (parentId)), (/**
                     * @param {?} entity
                     * @return {?}
                     */
                    function (entity) {
                        var _a;
                        return (/** @type {?} */ ((_a = {},
                            _a[key] = arrayUpdate((/** @type {?} */ (entity[key])), id, data, idKey),
                            _a)));
                    }));
                }
            }
        };
        try {
            for (var actions_1 = __values(actions), actions_1_1 = actions_1.next(); !actions_1_1.done; actions_1_1 = actions_1.next()) {
                var action = actions_1_1.value;
                _loop_1(action);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (actions_1_1 && !actions_1_1.done && (_a = actions_1.return)) _a.call(actions_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
    });
    /**
     * Stay in sync a subquery of the collection
     * \@param subQuery The sub query based on the entity
     * \@param child An object that describe the parent
     * @type {?}
     */
    var syncSubQuery = (/**
     * @template K
     * @param {?} subQuery
     * @param {?} child
     * @return {?}
     */
    function (subQuery, child) {
        var _a;
        var parentId = child.parentId, key = child.key;
        // If it's a static value
        if (!isQuery(subQuery)) {
            /** @type {?} */
            var update = _this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ (_a = {}, _a[key] = subQuery, _a)));
            return of(update);
        }
        if (Array.isArray(subQuery)) {
            /** @type {?} */
            var syncQueries = subQuery.map((/**
             * @param {?} oneQuery
             * @return {?}
             */
            function (oneQuery) {
                if (isQuery(subQuery)) {
                    /** @type {?} */
                    var id_1 = getIdAndPath({ path: subQuery.path });
                    return _this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
                     * @param {?} childDoc
                     * @return {?}
                     */
                    function (childDoc) {
                        _this['store'].update((/** @type {?} */ (parentId)), (/**
                         * @param {?} entity
                         * @return {?}
                         */
                        function (entity) {
                            var _a;
                            return (/** @type {?} */ ((_a = {},
                                _a[key] = arrayAdd((/** @type {?} */ (entity[key])), id_1, childDoc),
                                _a)));
                        }));
                    })));
                }
                return syncSubQuery(oneQuery, child);
            }));
            return combineLatest(syncQueries);
        }
        if (typeof subQuery !== 'object') {
            throw new Error('Query should be either a path, a Query object or an array of Queries');
        }
        // Sync subquery
        if (isDocPath(subQuery.path)) {
            return _this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
             * @param {?} children
             * @return {?}
             */
            function (children) {
                var _a;
                return _this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ (_a = {}, _a[key] = children, _a)));
            })));
        }
        else {
            return (/** @type {?} */ (_this['db'].collection(subQuery.path, subQuery.queryFn)
                .stateChanges()
                .pipe(withTransaction((/**
             * @param {?} actions
             * @return {?}
             */
            function (actions) { return fromChildAction(actions, child); })))));
        }
    });
    /**
     * Get in sync with all the subqueries
     * \@param subQueries A map of all the subqueries
     * \@param parent The parent entity to attach the children to
     * @type {?}
     */
    var syncAllSubQueries = (/**
     * @param {?} subQueries
     * @param {?} parent
     * @return {?}
     */
    function (subQueries, parent) {
        /** @type {?} */
        var obs = Object.keys(subQueries)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        function (key) { return key !== 'path' && key !== 'queryFn'; }))
            .map((/**
         * @param {?} key
         * @return {?}
         */
        function (key) {
            /** @type {?} */
            var queryLike = getSubQuery(subQueries[key], parent);
            /** @type {?} */
            var child = { key: key, parentId: parent[_this.idKey] };
            return syncSubQuery(queryLike, child);
        }));
        return combineLatest(obs);
    });
    ////////////////
    // MAIN QUERY //
    ////////////////
    /**
     * Listen on action with child queries
     * @type {?}
     */
    var fromActionWithChild = (/**
     * @param {?} actions
     * @param {?} mainQuery
     * @param {?} subscriptions
     * @return {?}
     */
    function (actions, mainQuery, subscriptions) {
        var e_2, _a, _b;
        try {
            for (var actions_2 = __values(actions), actions_2_1 = actions_2.next(); !actions_2_1.done; actions_2_1 = actions_2.next()) {
                var action = actions_2_1.value;
                /** @type {?} */
                var id = action.payload.doc.id;
                /** @type {?} */
                var data = action.payload.doc.data();
                switch (action.type) {
                    case 'added': {
                        /** @type {?} */
                        var entity = __assign((_b = {}, _b[_this.idKey] = id, _b), data);
                        _this['store'].upsert(id, entity);
                        subscriptions[id] = syncAllSubQueries((/** @type {?} */ (mainQuery)), entity).subscribe();
                        break;
                    }
                    case 'removed': {
                        _this['store'].remove(id);
                        subscriptions[id].unsubscribe();
                        delete subscriptions[id];
                        break;
                    }
                    case 'modified': {
                        _this['store'].update(id, data);
                    }
                }
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (actions_2_1 && !actions_2_1.done && (_a = actions_2.return)) _a.call(actions_2);
            }
            finally { if (e_2) throw e_2.error; }
        }
    });
    var path = query.path, queryFn = query.queryFn;
    if (isDocPath(path)) {
        var id_2 = this['getIdAndPath']({ path: path }).id;
        /** @type {?} */
        var subscription_1;
        return this['db'].doc(path).valueChanges().pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        function (entity) {
            _this['store'].upsert(id_2, (/** @type {?} */ (__assign({ id: id_2 }, entity))));
            if (!subscription_1) { // Subscribe only the first time
                subscription_1 = syncAllSubQueries((/** @type {?} */ (query)), entity).subscribe();
            }
        })), 
        // Stop subscription
        finalize((/**
         * @return {?}
         */
        function () { return subscription_1.unsubscribe(); })));
    }
    else {
        /** @type {?} */
        var subscriptions_1 = {};
        return (/** @type {?} */ (this['db'].collection(path, queryFn)
            .stateChanges()
            .pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        function (actions) { return fromActionWithChild(actions, query, subscriptions_1); })), 
        // Stop all subscriptions
        finalize((/**
         * @return {?}
         */
        function () { return Object.keys(subscriptions_1).forEach((/**
         * @param {?} id
         * @return {?}
         */
        function (id) {
            subscriptions_1[id].unsubscribe();
            delete subscriptions_1[id];
        })); })))));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Redirect the
 * @template E
 * @param {?} redirectTo Route path to redirecto if collection is empty
 * @return {?}
 */
function redirectIfEmpty(redirectTo) {
    return map((/**
     * @param {?} actions
     * @return {?}
     */
    function (actions) { return actions.length === 0 ? redirectTo : true; }));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { CollectionConfig, CollectionGroupService, CollectionGuard, CollectionGuardConfig, CollectionService, FireAuthService, SubcollectionService, canRead, canWrite, collection, doc, fireAuthProviders, getIdAndPath, getPathParams, hasRole, initialAuthState, isDocPath, pathWithParams, redirectIfEmpty, removeStoreEntity, setActive, setLoading, shoudCancel, syncQuery, syncStoreFromDocAction, syncStoreFromDocActionSnapshot, updateStoreEntity, upsertStoreEntity, waitForCancel };
//# sourceMappingURL=akita-ng-fire.js.map
